# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# See include/mesos/scheduler.hpp, include/mesos/executor.hpp and
# include/mesos/mesos.proto for more information documenting this
# interface.

"""Python bindings for Mesos."""

from __future__ import print_function

import sys

__all__ = (
  'Executor',
  'ExecutorDriver',
  'Scheduler',
  'SchedulerDriver',
  'OperatorMasterDriver',
  'OperatorMaster',
  'OperatorAgentDriver',
)

class Scheduler(object):
  """
    Base class for Mesos schedulers. Users' schedulers should extend this
    class to get default implementations of methods they don't override.
  """

  def registered(self, driver, frameworkId, masterInfo):
    """
      Invoked when the scheduler successfully registers with a Mesos master.
      It is called with the frameworkId, a unique ID generated by the
      master, and the masterInfo which is information about the master
      itself.
    """

  def reregistered(self, driver, masterInfo):
    """
      Invoked when the scheduler re-registers with a newly elected Mesos
      master.  This is only called when the scheduler has previously been
      registered.  masterInfo contains information about the newly elected
      master.
    """

  def disconnected(self, driver):
    """
      Invoked when the scheduler becomes disconnected from the master, e.g.
      the master fails and another is taking over.
    """

  def processHeartBeat(self, driver):
    """
      Invoked when the scheduler gets a heartbeat
    """

  def resourceOffers(self, driver, offers):
    """
      Invoked when resources have been offered to this framework. A single
      offer will only contain resources from a single slave.  Resources
      associated with an offer will not be re-offered to _this_ framework
      until either (a) this framework has rejected those resources (see
      SchedulerDriver.launchTasks) or (b) those resources have been
      rescinded (see Scheduler.offerRescinded).  Note that resources may be
      concurrently offered to more than one framework at a time (depending
      on the allocator being used).  In that case, the first framework to
      launch tasks using those resources will be able to use them while the
      other frameworks will have those resources rescinded (or if a
      framework has already launched tasks with those resources then those
      tasks will fail with a TASK_LOST status and a message saying as much).
    """

  def inverseOffers(self, driver, offers):
    """
      Invoked when an inverse offer is sent to the framework.  An inverse
      offer and a resource offer can hold many of the same fields, but an
      inverse offer requests resources rather than offering them.  Inverse
      offers may be accepted, rejected, re-offered, and rescinded.
    """

  def offerRescinded(self, driver, offerId):
    """
      Invoked when an offer is no longer valid (e.g., the slave was lost or
      another framework used resources in the offer.) If for whatever reason
      an offer is never rescinded (e.g., dropped message, failing over
      framework, etc.), a framework that attempts to launch tasks using an
      invalid offer will receive TASK_LOST status updates for those tasks
      (see Scheduler.resourceOffers).
    """

  def inverseOfferRescinded(self, driver, offerId):
    """
      Invoked when an inverse offer is no longer valid (e.g., the slave was lost or
      another framework used resources in the inverse offer.) If for whatever reason
      an inverse offer is never rescinded (e.g., dropped message, failing over
      framework, etc.), a framework that attempts to launch tasks using an
      invalid inverse offer will receive TASK_LOST status updates for those tasks
      (see Scheduler.resourceOffers).
    """

  def statusUpdate(self, driver, status):
    """
      Invoked when the status of a task has changed (e.g., a slave is
      lost and so the task is lost, a task finishes and an executor
      sends a status update saying so, etc). If implicit
      acknowledgements are being used, then returning from this
      callback _acknowledges_ receipt of this status update! If for
      whatever reason the scheduler aborts during this callback (or
      the process exits) another status update will be delivered (note,
      however, that this is currently not true if the slave sending the
      status update is lost/fails during that time). If explicit
      acknowledgements are in use, the scheduler must acknowledge this
      status on the driver.
    """

  def operationStatusUpdate(self, driver, status):
    """
      Invoked when there is an operation status update generated by the
      master, agent, or resource provider. These updates are only sent to
      the framework for operations which had the operation ID set by the
      framework. It is the responsibility of the scheduler to explicitly
      acknowledge the receipt of a status update.
    """

  def frameworkMessage(self, driver, executorId, slaveId, message):
    """
      Invoked when an executor sends a message. These messages are best
      effort; do not expect a framework message to be retransmitted in any
      reliable fashion.
    """

  def slaveLost(self, driver, slaveId):
    """
      Invoked when a slave has been determined unreachable (e.g., machine
      failure, network partition.) Most frameworks will need to reschedule
      any tasks launched on this slave on a new slave.

      NOTE: This callback is not reliably delivered. If a host or
      network failure causes messages between the master and the
      scheduler to be dropped, this callback may not be invoked.
    """

  def executorLost(self, driver, executorId, slaveId, status):
    """
      Invoked when an executor has exited/terminated. Note that any tasks
      running will have TASK_LOST status updates automatically generated.

      NOTE: This callback is not reliably delivered. If a host or
      network failure causes messages between the master and the
      scheduler to be dropped, this callback may not be invoked.
    """

  def error(self, driver, message):
    """
      Invoked when there is an unrecoverable error in the scheduler or
      scheduler driver.  The driver will be aborted BEFORE invoking this
      callback.
    """
    print("Error from Mesos: %s " % message, file=sys.stderr)


class SchedulerDriver(object):
  """
    Interface for Mesos scheduler drivers. Users may wish to implement this
    class in mock objects for tests.
  """
  def start(self):
    """
      Starts the scheduler driver.  This needs to be called before any other
      driver calls are made.
    """

  def stop(self, failover=False):
    """
      Stops the scheduler driver. If the 'failover' flag is set to False
      then it is expected that this framework will never reconnect to Mesos
      and all of its executors and tasks can be terminated.  Otherwise, all
      executors and tasks will remain running (for some framework specific
      failover timeout) allowing the scheduler to reconnect (possibly in the
      same process, or from a different process, for example, on a different
      machine.)
    """

  def abort(self):
    """
      Aborts the driver so that no more callbacks can be made to the
      scheduler.  The semantics of abort and stop have deliberately been
      separated so that code can detect an aborted driver (i.e., via the
      return status of SchedulerDriver.join), and instantiate and start
      another driver if desired (from within the same process.)
    """

  def join(self):
    """
      Waits for the driver to be stopped or aborted, possibly blocking the
      current thread indefinitely.  The return status of this function can
      be used to determine if the driver was aborted (see mesos.proto for a
      description of Status).
    """

  def run(self):
    """
      Starts and immediately joins (i.e., blocks on) the driver.
    """

  def requestResources(self, requests):
    """
      Requests resources from Mesos (see mesos.proto for a description of
      Request and how, for example, to request resources from specific
      slaves.)  Any resources available are offered to the framework via
      Scheduler.resourceOffers callback, asynchronously.
    """

  def launchTasks(self, offerIds, tasks, filters=None):
    """
      Launches the given set of tasks. Any remaining resources (i.e.,
      those that are not used by the launched tasks or their executors)
      will be considered declined. Note that this includes resources
      used by tasks that the framework attempted to launch but failed
      (with TASK_ERROR) due to a malformed task description. The
      specified filters are applied on all unused resources (see
      mesos.proto for a description of Filters). Available resources
      are aggregated when multiple offers are provided. Note that all
      offers must belong to the same slave. Invoking this function with
      an empty collection of tasks declines offers in their entirety
      (see Scheduler.declineOffer). Note that passing a single offer
      is also supported.
    """

  def killTask(self, taskId):
    """
      Kills the specified task. Note that attempting to kill a task is
      currently not reliable.  If, for example, a scheduler fails over while
      it was attempting to kill a task it will need to retry in the future.
      Likewise, if unregistered / disconnected, the request will be dropped
      dropped (these semantics may be changed in the future).
    """

  def acceptOffers(self, offerIds, operations, filters=None):
    """
      Accepts the given offers and performs a sequence of operations on
      those accepted offers. See Offer.Operation in mesos.proto for the
      set of available operations. Any remaining resources (i.e., those
      that are not used by the launched tasks or their executors) will
      be considered declined. Note that this includes resources used by
      tasks that the framework attempted to launch but failed (with
      TASK_ERROR) due to a malformed task description. The specified
      filters are applied on all unused resources (see mesos.proto for
      a description of Filters). Available resources are aggregated
      when multiple offers are provided. Note that all offers must
      belong to the same slave.
    """

  def acceptInverseOffers(self, offer_ids, filters=None):
    """
      Accepts an inverse offer. Inverse offers should only be accepted
      if the resources in the offer can be safely evacuated before the
      provided unavailability.
    """

  def declineOffer(self, offerId, filters=None):
    """
      Declines an offer in its entirety and applies the specified
      filters on the resources (see mesos.proto for a description of
      Filters). Note that this can be done at any time, it is not
      necessary to do this within the Scheduler::resourceOffers
      callback.
    """

  def declineInverseOffer(self, offer_ids, filters=None):
    """
      Declines an inverse offer. Inverse offers should be declined if
      the resources in the offer might not be safely evacuated before
      the provided unavailability.
    """

  def reviveOffers(self):
    """
      Removes all filters previously set by the framework (via
      launchTasks()).  This enables the framework to receive offers from
      those filtered slaves.
    """

  def suppressOffers(self):
    """
      Inform Mesos master to stop sending offers to the framework. The
      scheduler should call reviveOffers() to resume getting offers.
    """

  def acknowledgeStatusUpdate(self, status):
    """
      Acknowledges the status update. This should only be called
      once the status update is processed durably by the scheduler.
      Not that explicit acknowledgements must be requested via the
      constructor argument, otherwise a call to this method will
      cause the driver to crash.
    """

  def acknowledgeOperationStatusUpdate(self, status):
    """
      Acknowledges the operation status update. This should only be called
      once the operation status update is processed durably by the scheduler.
      Not that explicit acknowledgements must be requested via the
      constructor argument, otherwise a call to this method will
      cause the driver to crash.
    """

  def sendFrameworkMessage(self, executorId, slaveId, data):
    """
      Sends a message from the framework to one of its executors. These
      messages are best effort; do not expect a framework message to be
      retransmitted in any reliable fashion.
    """

  def reconcileTasks(self, tasks):
    """
      Allows the framework to query the status for non-terminal tasks.
      This causes the master to send back the latest task status for
      each task in 'statuses', if possible. Tasks that are no longer
      known will result in a TASK_LOST update. If statuses is empty,
      then the master will send the latest status for each task
      currently known.
    """

class Executor(object):
  """
    Base class for Mesos executors. Users' executors should extend this
    class to get default implementations of methods they don't override.
  """

  def registered(self, driver, executorInfo, frameworkInfo, slaveInfo):
    """
      Invoked once the executor driver has been able to successfully connect
      with Mesos.  In particular, a scheduler can pass some data to its
      executors through the FrameworkInfo.ExecutorInfo's data field.
    """

  def reregistered(self, driver, slaveInfo):
    """
      Invoked when the executor re-registers with a restarted slave.
    """

  def disconnected(self, driver):
    """
      Invoked when the executor becomes "disconnected" from the slave (e.g.,
      the slave is being restarted due to an upgrade).
    """

  def launchTask(self, driver, task):
    """
      Invoked when a task has been launched on this executor (initiated via
      Scheduler.launchTasks).  Note that this task can be realized with a
      thread, a process, or some simple computation, however, no other
      callbacks will be invoked on this executor until this callback has
      returned.
    """

  def launchTaskGroup(self, driver, task_infos):
    """
      Invoked when a task group has been launched on this executor (initiated via
      Scheduler.launchTasks).  Note that this task can be realized with a
      thread, a process, or some simple computation, however, no other
      callbacks will be invoked on this executor until this callback has
      returned.
    """

  def killTask(self, driver, taskId):
    """
      Invoked when a task running within this executor has been killed (via
      SchedulerDriver.killTask).  Note that no status update will be sent on
      behalf of the executor, the executor is responsible for creating a new
      TaskStatus (i.e., with TASK_KILLED) and invoking ExecutorDriver's
      sendStatusUpdate.
    """

  def frameworkMessage(self, driver, message):
    """
      Invoked when a framework message has arrived for this executor.  These
      messages are best effort; do not expect a framework message to be
      retransmitted in any reliable fashion.
    """

  def shutdown(self, driver):
    """
      Invoked when the executor should terminate all of its currently
      running tasks.  Note that after Mesos has determined that an executor
      has terminated any tasks that the executor did not send terminal
      status updates for (e.g., TASK_KILLED, TASK_FINISHED, TASK_FAILED,
      etc) a TASK_LOST status update will be created.
    """

  def error(self, driver, message):
    """
      Invoked when a fatal error has occurred with the executor and/or
      executor driver.  The driver will be aborted BEFORE invoking this
      callback.
    """
    print("Error from Mesos: %s" % message, file=sys.stderr)



class ExecutorDriver(object):
  """
    Interface for Mesos executor drivers. Users may wish to extend this
    class in mock objects for tests.
  """
  def start(self):
    """
      Starts the executor driver. This needs to be called before any other
      driver calls are made.
    """

  def stop(self):
    """
      Stops the executor driver.
    """

  def abort(self):
    """
      Aborts the driver so that no more callbacks can be made to the
      executor.  The semantics of abort and stop have deliberately been
      separated so that code can detect an aborted driver (i.e., via the
      return status of ExecutorDriver.join), and instantiate and start
      another driver if desired (from within the same process, although this
      functionality is currently not supported for executors).
    """

  def join(self):
    """
      Waits for the driver to be stopped or aborted, possibly blocking the
      current thread indefinitely.  The return status of this function can
      be used to determine if the driver was aborted (see mesos.proto for a
      description of Status).
    """

  def run(self):
    """
      Starts and immediately joins (i.e., blocks on) the driver.
    """

  def sendStatusUpdate(self, status):
    """
      Sends a status update to the framework scheduler, retrying as
      necessary until an acknowledgement has been received or the executor
      is terminated (in which case, a TASK_LOST status update will be sent).
      See Scheduler.statusUpdate for more information about status update
      acknowledgements.
    """

  def sendFrameworkMessage(self, data):
    """
      Sends a message to the framework scheduler. These messages are best
      effort; do not expect a framework message to be retransmitted in any
      reliable fashion.
    """


class OperatorDaemonDriver(object):
  """
    Operator HTTP API: Operations common to master daemon and agent daemon.
  """

  def getHealth(self):
    """
      This call retrieves the health status of master daemon or agent daemon.
    """

  def getFlags(self):
    """
      This call retrieves the daemon's overall flag configuration.
    """

  def getVersion(self):
    """
      This call retrieves the daemon's version information.
    """

  def getMetrics(self, timeout):
    """
      This call gives the snapshot of current metrics to the end user. If timeout is set in the call, it would be used
      to determine the maximum amount of time the API will take to respond. If the timeout is exceeded, some metrics may
      not be included in the response.
    """

  def getLoggingLevel(self):
    """
      This call retrieves the daemon's logging level.
    """

  def setLoggingLevel(self, level, duration):
    """
      Sets the logging verbosity level for a specified duration for master daemon or agent daemon. Mesos uses glog for
      logging. The library only uses verbose logging which means nothing will be output unless the verbosity level is
      set (by default it's 0, libprocess uses levels 1, 2, and 3).
    """

  def listFiles(self, path):
    """
      This call retrieves the file listing for a directory in master daemon or agent daemon.
    """

  def readFile(self, path, length, offset):
    """
      Reads data from a file. This call takes path of the file to be read in the daemon, offset to start reading
      position and length for the maximum number of bytes to read.
    """

  def getState(self):
    """
      This call retrieves the overall cluster state.
    """

  def getFrameworks(self):
    """
      This call retrieves information about all the frameworks known to the master daemon or agent daemon.
    """

  def getExecutors(self):
    """
      Queries about all the executors known to the master daemon or agent daemon.
    """

  def getTasks(self):
    """
      Query about all the tasks known to the master daemon or agent daemon.
    """


class OperatorMasterDriver(OperatorDaemonDriver):
  """
    Interface for Mesos operator drivers. Users may wish to extend this
    class in mock objects for tests.
  """

  def start(self):
    """
      Starts the operator driver. This needs to be called before any other
      driver calls are made.
    """

  def stop(self):
    """
      Stops the operator driver.
    """

  def abort(self):
    """
      Aborts the driver so that no more callbacks can be made to the
      operator.  The semantics of abort and stop have deliberately been
      separated so that code can detect an aborted driver (i.e., via the
      return status of OperatorMasterDriver.join), and instantiate and start
      another driver if desired (from within the same process, although this
      functionality is currently not supported for executors).
    """

  def join(self):
    """
      Waits for the driver to be stopped or aborted, possibly blocking the
      current thread indefinitely.  The return status of this function can
      be used to determine if the driver was aborted (see mesos.proto for a
      description of Status).
    """

  def run(self):
    """
      Starts and immediately joins (i.e., blocks on) the driver.
    """

  def getAgents(self):
    """
      This call retrieves information about all the agents known to the master.
    """

  def getRoles(self):
    """
      Query the information about roles.
    """

  def getWeights(self):
    """
      This call retrieves the information about role weights.
    """

  def updateWeights(self, weight_infos):
    """
      This call updates weights for specific role. This call takes weight_infos which needs role value and weight value.
    """

  def getMaster(self):
    """
      This call retrieves the information on master.
    """

  def reserveResources(self, agent_id, resources):
    """
      This call reserve resources dynamically on a specific agent. This call takes agent_id and resources details like
      the following.
    """

  def unreserveResources(self, agent_id, resources):
    """
      This call unreserve resources dynamically on a specific agent. This call takes agent_id and resources details like
      the following.
    """

  def createVolumes(self, agent_id, volumes):
    """
      This call create persistent volumes on reserved resources. The request is forwarded asynchronously to the Mesos
      agent where the reserved resources are located. That asynchronous message may not be delivered or creating the
      volumes at the agent might fail. This call takes agent_id and volumes details like the following.
    """

  def destroyVolumes(self, agent_id, volumes):
    """
      This call destroys persistent volumes. The request is forwarded asynchronously to the Mesos agent where the
      reserved resources are located.
    """

  def getMaintenanceStatus(self):
    """
      This call retrieves the cluster's maintenance status.
    """

  def getMaintenanceSchedule(self):
    """
      This call retrieves the cluster's maintenance status.
    """

  def updateMaintenanceSchedule(self, windows):
    """
      This call retrieves the cluster's maintenance schedule.
    """

  def startMaintenance(self, machines):
    """
      This call starts the maintenance of the cluster, this would bring a set of machines down.
    """

  def stopMaintenance(self, machines):
    """
      Stops the maintenance of the cluster, this would bring a set of machines back up.
    """

  def getQuota(self):
    """
      This call retrieves the cluster's configured quotas.
    """

  def setQuota(self, quota_request):
    """
      This call sets the quota for resources to be used by a particular role.
    """

  def removeQuota(self, role):
    """
      This call removes the quota for a particular role.
    """

  def markAgentGone(self, agent_id):
    """
      This call can be used by operators to assert that an agent instance has failed and is never coming back
      (e.g., ephemeral instance from cloud provider). The master would shutdown the agent and send TASK_GONE_BY_OPERATOR
      updates for all the running tasks. This signal can be used by stateful frameworks to re-schedule their workloads
      (volumes, reservations etc.) to other agent instances. It is possible that the tasks might still be running if the
      operator's assertion was wrong and the agent was partitioned away from the master. The agent would be shutdown
      when it tries to re-register with the master when the partition heals. This call is idempotent.
    """

class OperatorMaster(object):
  """
    Base class for Mesos operators. Users' operators should extend this
    class to get default implementations of methods they don't override.
  """

  # def heartbeat(self):
  #   """
  #     Periodically sent by the master to the subscriber according to 'Subscribed.heartbeat_interval_seconds'. If the
  #     subscriber does not receive any events (including heartbeats) for an extended period of time (e.g., 5 x
  #     heartbeat_interval_seconds), it is likely that the connection is lost or there is a network partition. In that
  #     case, the subscriber should close the existing subscription connection and resubscribe using a backoff strategy.
  #   """

  def taskAdded(self, task_info):
    """
      Invoked whenever a task has been added to the master. This can happen either when a new task launch is processed
      by the master or when an agent re-registers with a failed over master.
    """

  def taskUpdated(self, task_info):
    """
      Invoked whenever the state of the task changes in the master. This can happen when a status update is received or
      generated by the master. Since status updates are retried by the agent, not all status updates received by the
      master result in the event being sent.
    """

  def frameworkAdded(self, framework_info):
    """
      Sent whenever a framework becomes known to the master. This can happen when a new framework registers with the
      master.
    """

  def frameworkUpdated(self, framework_info):
    """
      Sent whenever a framework re-registers with the master upon a disconnection (network error) or upon a master
      failover.
    """

  def frameworkRemoved(self, framework_info):
    """
      Sent whenever a framework is removed. This can happen when a framework is explicitly teardown by the operator or
      if it fails to re-register with the master within the failover timeout.
    """

  def agentAdded(self, agent_info):
    """
      Sent whenever an agent becomes known to it. This can happen when an agent registered for the first time, or
      reregistered after a master failover.
    """

  def agentRemoved(self, agent_id):
    """
      Sent whenever a agent is removed. This can happen when the agent is scheduled for maintenance. (NOTE: It's
      possible that an agent might become active once it has been removed, i.e. if the master has gc'ed its list of
      known "dead" agents. See MESOS-5965 for context).
    """

class OperatorAgentDriver(OperatorDaemonDriver):
  """
    This API contains all the calls accepted by the agent.
  """

  def getContainers(self, show_nested=False, show_standalone=False):
    """
      This call retrieves information about containers running on this agent. It contains ContainerStatus and
      ResourceStatistics along with some metadata of the containers.
      There are two knobs in the request to control the types of the containers this API will return:
      * show_nested: Whether to show nested containers [default: false].
      * show_standalone: Whether to show standalone containers [default: false].
    """

  def launchNestedContainer(self, launch_nested_container):
    """
      This call launches a nested container. Any authorized entity, including the executor itself, its tasks, or the
      operator can use this API to launch a nested container.
    """

  def waitNestedContainer(self, container_id, parent_id=None):
    """
      This call waits for a nested container to terminate or exit. Any authorized entity, including the executor itself,
      its tasks, or the operator can use this API to wait on a nested container.
    """

  def killNestedContainer(self, container_id, parent_id=None):
    """
      This call initiates the destruction of a nested container. Any authorized entity, including the executor itself,
      its tasks, or the operator can use this API to kill a nested container.
    """

  def launchNestedContainerSession(self, launch_nested_container_session):
    """
      This call launches a nested container whose lifetime is tied to the lifetime of the HTTP call establishing this
      connection. The STDOUT and STDERR of the nested container is streamed back to the client so long as the connection
      is active.
    """

  def attachContainerInput(self, container_id):
    """
      This call attaches to the STDIN of the primary process of a container and streams input to it. This call can only
      be made against containers that have been launched with an associated IOSwitchboard (i.e. nested containers
      launched via a LAUNCH_NESTED_CONTAINER_SESSION call or normal containers launched with a TTYInfo in their
      ContainerInfo). Only one ATTACH_CONTAINER_INPUT call can be active for a given container at a time. Subsequent
      attempts to attach will fail.

      The first message sent over an ATTACH_CONTAINER_INPUT stream must be of type CONTAINER_ID and contain the
      ContainerID of the container being attached to. Subsequent messages must be of type PROCESS_IO, but they may
      contain subtypes of either DATA or CONTROL. DATA messages must be of type STDIN and contain the actual data to
      stream to the STDIN of the container being attached to. Currently, the only valid CONTROL message sends a
      heartbeat to keep the connection alive. We may add more CONTROL messages in the future.
    """

  def attachContainerOutput(self, container_id):
    """
      This call attaches to the STDOUT and STDERR of the primary process of a container and streams its output back to
      the client. This call can only be made against containers that have been launched with an associated IOSwitchboard
      (i.e. nested containers launched via a LAUNCH_NESTED_CONTAINER_SESSION call or normal containers launched with a
      TTYInfo in their ContainerInfo field). Multiple ATTACH_CONTAINER_OUTPUT calls can be active for a given container
      at once.
    """

  def removeNestedContainer(self, container_id, parent_id = None):
    """
      This call triggers the removal of a nested container and its artifacts (e.g., the sandbox and runtime directories).
      This call can only be made against containers that have already terminated, and whose parent container has not
      been destroyed. Any authorized entity, including the executor itself, its tasks, or the operator can use this API
      call.
    """

  def addResourceProviderConfig(self, info):
    """
      This call launches a Local Resource Provider on the agent with the specified ResourceProviderInfo.
    """

  def updateResourceProviderConfig(self, info):
    """
      This call updates a Local Resource Provider on the agent with the specified ResourceProviderInfo.
    """

  def removeResourceProviderConfig(self, type, name):
    """
      This call terminates a given Local Resource Provider on the agent and prevents it from being launched again until
      the config is added back. The master and the agent will think the resource provider has disconnected, similar to
      agent disconnection.
      If there exists a task that is using the resources provided by the resource provider, its execution will not be
      affected. However, offer operations for the local resource provider will not be successful. In fact, if a local
      resource provider is disconnected, the master will rescind the offers related to that local resource provider,
      effectively disallowing frameworks to perform operations on the disconnected local resource provider.
      The local resource provider can be re-added after its removal using ADD_RESOURCE_PROVIDER_CONFIG. Note that
      removing a local resource provider is different than marking a local resource provider as gone, in which case the
      local resource provider will not be allowed to be re-added. Marking a local resource provider as gone is not yet
      supported.
    """

  def pruneImages(self, excluded_images=None):
    """
      This call triggers garbage collection for container images. This call can only be made when all running containers
      are launched with Mesos version 1.5 or newer. An optional list of excluded images from GC can be speficied via
      prune_images.excluded_images field.
    """



import json
import logging
from threading import RLock

from addict import Dict
from six.moves.http_client import HTTPConnection
from six.moves.urllib.parse import urlparse

from .interface import OperatorDaemonDriver
from .process import Process
from .utils import DAY

logger = logging.getLogger(__name__)


class MesosOperatorDaemonDriver(OperatorDaemonDriver):
    _timeout = 10

    def __init__(self, daemon_uri):
        self.init(daemon_uri)

    def init(self, daemon_uri):
        """
        :param daemon_uri: masterHost:5050 or agentHost:5051
        """
        self._daemon = daemon_uri
        self._conn = None

    def _get_conn(self):
        if self._conn is not None:
            return self._conn

        host, port = self._daemon.split(':', 2)
        port = int(port)
        self._conn = HTTPConnection(host, port, timeout=self._timeout)
        return self._conn

    def _send(self, body, path='/api/v1/operator', method='POST', headers={}):
        with self._lock:
            conn = self._get_conn()
            if conn is None:
                raise RuntimeError('Not connected yet')

            if body != '':
                data = json.dumps(body).encode('utf-8')
                headers['Content-Type'] = 'application/json'
            else:
                data = ''

            # stream_id = self.stream_id
            # if stream_id:
            #     headers['Mesos-Stream-Id'] = stream_id

            try:
                conn.request(method, path, body=data, headers=headers)
                resp = conn.getresponse()
            except Exception:
                self._conn.close()
                self._conn = None
                raise

            if resp.status >= 300 and resp.status <= 399:
                url = resp.getheader('location')
                parsed = urlparse(url)
                self._daemon = '%s:%s' % (parsed.hostname, parsed.port)
                self._conn.close()
                self._conn = None
                return self._send(body, path, method, headers)

            if resp.status < 200 or resp.status >= 300:
                raise RuntimeError(
                    'Failed to send request code=%s, message=%s' % (
                        resp.status, resp.read()
                    )
                )

            result = resp.read()
            if not result:
                return {}

            try:
                return json.loads(result.decode('utf-8'))
            except Exception:
                return {}

    def getHealth(self):
        body = dict(
            type='GET_HEALTH',
        )
        return self._send(body)

    def getFlags(self):
        body = dict(
            type='GET_FLAGS',
        )
        return self._send(body)

    def getVersion(self):
        body = dict(
            type='GET_VERSION',
        )
        return self._send(body)

    def getMetrics(self, timeout):
        body = dict(
            type='GET_METRICS',
            get_metrics=dict(
                timeout=dict(
                    nanoseconds=timeout,
                )
            )
        )
        return self._send(body)

    def getLoggingLevel(self):
        body = dict(
            type='GET_LOGGING_LEVEL',
        )
        return self._send(body)

    def setLoggingLevel(self, level, duration):
        body = dict(
            type='SET_LOGGING_LEVEL',
            set_logging_level=dict(
                duration=dict(
                    nanoseconds=duration,
                ),
                level=level,
            )
        )
        return self._send(body)

    def listFiles(self, path):
        body = dict(
            type='LIST_FILES',
            list_files=dict(
                path=path,
            ),
        )
        return self._send(body)

    def readFile(self, path, length, offset):
        body = dict(
            type='READ_FILE',
            read_file=dict(
                length=length,
                offset=offset,
                path=path,
            )
        )
        return self._send(body)

    def getState(self):
        body = dict(
            type='GET_STATE',
        )
        return self._send(body)

    def getFrameworks(self):
        body = dict(
            type='GET_FRAMEWORKS',
        )
        return self._send(body)

    def getExecutors(self):
        body = dict(
            type='GET_EXECUTORS',
        )
        return self._send(body)

    def getTasks(self):
        body = dict(
            type='GET_TASKS',
        )
        return self._send(body)


class MesosOperatorMasterDriver(Process, MesosOperatorDaemonDriver):
    def __init__(self, master_uri, operator=None, use_addict=False,
                 timeout=DAY):
        """

        :param master_uri:
        :param operator: Optional. Only if you want to send requests that
        result in a stream of events (SUBSCRIBE).
        :type operator: OperatorMaster
        """
        super(MesosOperatorMasterDriver, self).__init__(master=master_uri,
                                                        timeout=timeout)
        super(MesosOperatorMasterDriver, self).init(master_uri)
        self.operator = operator
        self.master_uri = master_uri
        self._dict_cls = Dict if use_addict else dict

    def start(self):
        super(MesosOperatorMasterDriver, self).start()
        uri = self.master_uri
        if uri.startswith('zk://') or uri.startswith('zoo://'):
            from .detector import MasterDetector
            self.detector = MasterDetector(uri[uri.index('://') + 3:], self)
            self.detector.start()
        else:
            if ':' not in uri:
                uri += ':5050'
            self.change_master(uri)

    def gen_request(self):
        body = json.dumps(
            dict(
                type='SUBSCRIBE',
            ),
        )

        request = ('POST /api/v1/operator HTTP/1.1\r\nHost: %s\r\n'
                   'Content-Type: application/json\r\n'
                   'Accept: application/json\r\n'
                   'Connection: close\r\nContent-Length: %s\r\n\r\n%s') % (
                      self.master, len(body), body
                  )
        return request.encode('utf-8')

    def on_event(self, event):
        if 'type' in event:
            _type = event['type'].lower()
            if _type == 'heartbeat':
                return

            if _type not in event:
                logger.error(
                    'Missing `%s` in event %s' %
                    (_type, event))
                return

            event = event[_type]
            func_name = 'on_%s' % (_type,)
            func = getattr(self, func_name, None)
            if func is not None:
                func(event)
            else:
                logger.error('Unknown type:%s, event:%s' % (_type, event))
        else:
            logger.error('Unknown event:%s' % (event,))

    def on_close(self):
        if self._conn is not None:
            self._conn.close()
            self._conn = None

    def on_subscribed(self, info):
        state = info['get_state']
        logger.info(
            'Operator client subscribed with cluster state: %s' % state)

    def on_task_added(self, event):
        task_info = event['task']
        self.operator.taskAdded(self._dict_cls(task_info))

    def on_task_updated(self, event):
        self.operator.taskUpdated(self._dict_cls(event))

    def on_framework_added(self, event):
        framework_info = event['framework']
        self.operator.frameworkAdded(self._dict_cls(framework_info))

    def on_framework_updated(self, event):
        framework_info = event['framework']
        self.operator.frameworkUpdated(self._dict_cls(framework_info))

    def on_framework_removed(self, event):
        framework_info = event['framework_info']
        self.operator.frameworkRemoved(self._dict_cls(framework_info))

    def on_agent_added(self, event):
        agent_info = event['agent']
        self.operator.agentAdded(self._dict_cls(agent_info))

    def on_agent_removed(self, event):
        agent_id = event['agent_id']['value']
        self.operator.agentRemoved(agent_id)

    def getAgents(self):
        body = dict(
            type='GET_AGENTS',
        )
        return self._send(body)

    def getRoles(self):
        body = dict(
            type='GET_ROLES',
        )
        return self._send(body)

    def getWeights(self):
        body = dict(
            type='GET_WEIGHTS',
        )
        return self._send(body)

    def updateWeights(self, weight_infos):
        body = dict(
            type='UPDATE_WEIGHTS',
            update_weights=dict(
                weight_infos=[dict(role=weight_info['role'],
                                   weight=weight_info['weight']) for
                              weight_info in
                              weight_infos],
            ),
        )
        self._send(body)

    def getMaster(self):
        body = dict(
            type='GET_MASTER',
        )
        return self._send(body)

    def reserveResources(self, agent_id, resources):
        body = dict(
            type='RESERVE_RESOURCES',
            reserve_resources=dict(
                agent_id=dict(value=agent_id),
                resources=resources,
            ),
        )
        self._send(body)

    def unreserveResources(self, agent_id, resources):
        body = dict(
            type='UNRESERVE_RESOURCES',
            unreserve_resources=dict(
                agent_id=dict(value=agent_id),
                resources=resources,
            ),
        )
        self._send(body)

    def createVolumes(self, agent_id, volumes):
        body = dict(
            type='CREATE_VOLUMES',
            create_volumes=dict(
                agent_id=dict(value=agent_id),
                volumes=volumes,
            ),
        )
        self._send(body)

    def destroyVolumes(self, agent_id, volumes):
        body = dict(
            type='DESTROY_VOLUMES',
            destroy_volumes=dict(
                agent_id=dict(value=agent_id),
                volumes=volumes,
            ),
        )
        self._send(body)

    def getMaintenanceStatus(self):
        body = dict(
            type='GET_MAINTENANCE_STATUS',
        )
        return self._send(body)

    def getMaintenanceSchedule(self):
        body = dict(
            type='GET_MAINTENANCE_SCHEDULE',
        )
        return self._send(body)

    def updateMaintenanceSchedule(self, windows):
        body = dict(
            type='UPDATE_MAINTENANCE_SCHEDULE',
            update_maintenance_schedule=dict(
                schedule=dict(
                    windows=windows
                ),
            ),
        )
        self._send(body)

    def startMaintenance(self, machines):
        body = dict(
            type='START_MAINTENANCE',
            start_maintenance=dict(
                machines=[dict(
                    hostname=machine['hostname'], ip=machine['ip']) for machine
                    in machines],
            ),
        )
        self._send(body)

    def stopMaintenance(self, machines):
        body = dict(
            type='STOP_MAINTENANCE',
            stop_maintenance=dict(
                machines=[dict(
                    hostname=machine['hostname'], ip=machine['ip']) for machine
                    in machines],
            ),
        )
        self._send(body)

    def getQuota(self):
        body = dict(
            type='GET_QUOTA',
        )
        return self._send(body)

    def setQuota(self, quota_request):
        body = dict(
            type='SET_QUOTA',
            set_quota=dict(
                quota_request=quota_request,
            ),
        )
        self._send(body)

    def removeQuota(self, role):
        body = dict(
            type='REMOVE_QUOTA',
            remove_quota=dict(
                role=role,
            ),
        )
        self._send(body)

    def markAgentGone(self, agent_id):
        body = dict(
            type='MARK_AGENT_GONE',
            mark_agent_gone=dict(
                agent_id=dict(
                    value=agent_id,
                ),
            ),
        )
        self._send(body)


class MesosOperatorAgentDriver(MesosOperatorDaemonDriver):
    def __init__(self, agent_uri):
        super(MesosOperatorAgentDriver, self).__init__(agent_uri)
        self._lock = RLock()

    def getContainers(self, show_nested=False, show_standalone=False):
        body = dict(
            type='GET_CONTAINERS',
            get_containers=dict(
                show_nested=show_nested,
                show_standalone=show_standalone,
            ),
        )
        return self._send(body)

    def launchNestedContainer(self, launch_nested_container):
        body = dict(
            type='LAUNCH_NESTED_CONTAINER',
            launch_nested_container=launch_nested_container,
        )
        self._send(body)

    def waitNestedContainer(self, container_id, parent_id=None):
        body = dict(
            type='WAIT_NESTED_CONTAINER',
            wait_nested_container=dict(
                container_id=dict(
                    value=container_id,
                ),
            ),
        )
        if parent_id is not None:
            body['wait_nested_container']['container_id']['parent'] = dict(
                value=parent_id, )
        self._send(body)

    def killNestedContainer(self, container_id, parent_id=None):
        body = dict(
            type='KILL_NESTED_CONTAINER',
            kill_nested_container=dict(
                container_id=dict(
                    value=container_id,
                ),
            ),
        )
        if parent_id is not None:
            body['kill_nested_container']['container_id']['parent'] = dict(
                value=parent_id, )
        self._send(body)

    def launchNestedContainerSession(self, launch_nested_container_session):
        body = dict(
            type='LAUNCH_NESTED_CONTAINER_SESSION',
            launch_nested_container_session=launch_nested_container_session,
        )
        headers = {'Accept': 'application/recordio'}
        self._send(body, headers=headers)

    def attachContainerInput(self, container_id, process_ios):
        # The first message sent over an ATTACH_CONTAINER_INPUT stream must be
        # of type CONTAINER_ID and contain the ContainerID of the container
        # being attached to.
        msg = dict(
            type='ATTACH_CONTAINER_INPUT',
            attach_container_input=dict(
                type='CONTAINER_ID',
                container_id=dict(
                    value=container_id,
                )
            ),
        )
        # Messages are encoded in RecordIO format
        sort_keys = True
        msg_str = json.dumps(msg, sort_keys=sort_keys)
        record_io = '{}\n{}'.format(len(msg_str), msg_str)
        body = ''
        body += record_io
        # Subsequent messages must be of type PROCESS_IO
        # Template for PROCESS_IO messages
        process_io = dict(
            type='ATTACH_CONTAINER_INPUT',
            attach_container_input=dict(
                type='PROCESS_IO',
                process_io='DUMMY',
            )
        )
        for msg in process_ios:
            if msg['type'] != 'DATA' and msg['type'] != 'CONTROL':
                raise ValueError(
                    'PROCESS_IO messages may contain subtypes of either DATA \
                    or CONTROL')
            process_io['attach_container_input']['process_io'] = msg
            msg_str = json.dumps(process_io, sort_keys=sort_keys)
            record_io = '{}\n{}'.format(len(msg_str), msg_str)
            body += record_io
        headers = {'Accept': 'application/recordio'}
        self._send(body, headers=headers)

    def attachContainerOutput(self, container_id):
        body = dict(
            type='ATTACH_CONTAINER_OUTPUT',
            attach_container_output=dict(
                container_id=dict(
                    value=container_id,
                )
            ),
        )
        headers = {'Accept': 'application/recordio',
                   'Message-Accept': 'application/json'}
        return self._send(body, headers=headers)

    def removeNestedContainer(self, container_id, parent_id=None):
        body = dict(
            type='REMOVE_NESTED_CONTAINER',
            remove_nested_container=dict(
                container_id=dict(
                    value=container_id,
                ),
            ),
        )
        if parent_id is not None:
            body['remove_nested_container']['container_id']['parent'] = dict(
                value=parent_id, )
        self._send(body)

    def addResourceProviderConfig(self, info):
        body = dict(
            type='ADD_RESOURCE_PROVIDER_CONFIG',
            add_resource_provider_config=dict(
                info=info,
            ),
        )
        self._send(body)

    def updateResourceProviderConfig(self, info):
        body = dict(
            type='UPDATE_RESOURCE_PROVIDER_CONFIG',
            update_resource_provider_config=dict(
                info=info,
            ),
        )
        self._send(body)

    def removeResourceProviderConfig(self, type, name):
        body = dict(
            type='REMOVE_RESOURCE_PROVIDER_CONFIG',
            remove_resource_provider_config=dict(
                type=type,
                name=name,
            ),
        )
        self._send(body)

    def pruneImages(self, excluded_images=None):
        body = dict(
            type='PRUNE_IMAGES',
        )
        if excluded_images is not None:
            body['prune_images'] = dict(
                excluded_images=excluded_images,
            )
        self._send(body)



import sys
import logging

from .interface import Scheduler, Executor, OperatorMaster
from .scheduler import MesosSchedulerDriver
from .executor import MesosExecutorDriver
from .operator_v1 import MesosOperatorMasterDriver, MesosOperatorAgentDriver
from .utils import encode_data, decode_data


LOG_FMT = '%(asctime)-15s [%(levelname)s] [%(threadName)s]' \
      ' [%(name)-9s:%(lineno)d] %(message)s'
DATE_FMT = '%Y-%m-%d %H:%M:%S'

logger = logging.getLogger('pymesos')
handler = logging.StreamHandler(stream=sys.stderr)
handler.setFormatter(logging.Formatter(LOG_FMT, DATE_FMT))
logger.addHandler(handler)
logger.setLevel(logging.INFO)


__VERSION__ = '0.3.15'

__all__ = (
    'Scheduler',
    'MesosSchedulerDriver',
    'Executor',
    'MesosExecutorDriver',
    'encode_data',
    'decode_data',
    'OperatorMaster',
    'MesosOperatorMasterDriver',
    'MesosOperatorAgentDriver',
)



import os
import re
import sys
import json
import time
import math
import errno
import random
import select
import signal
import socket
import logging
from six import reraise
from six.moves import _thread as thread
from six.moves.http_client import OK, TEMPORARY_REDIRECT, SERVICE_UNAVAILABLE
from six.moves.urllib.parse import urlparse
from threading import Thread, RLock
from http_parser.http import HttpParser
from .utils import DAY


def _strerror(err):
    try:
        return os.strerror(err)
    except (ValueError, OverflowError, NameError):
        if err in errno.errorcode:
            return errno.errorcode[err]
        return "Unknown error %s" % err


def _handle_sigint(signum, frame):
    global _prev_handler, _exc_info
    assert signum == signal.SIGINT
    if _exc_info is not None:
        exc_info = _exc_info
        _exc_info = None
        reraise(*exc_info)
    elif _prev_handler is not None:
        return _prev_handler(signum, frame)

    raise KeyboardInterrupt


_exc_info = None
_prev_handler = signal.signal(signal.SIGINT, _handle_sigint)
LENGTH_PATTERN = re.compile(br'\d+\n')
logger = logging.getLogger(__name__)
PIPE_BUF = getattr(select, 'PIPE_BUF', 4096)

SLEEP_TIMEOUT_SCALE = 2
SLEEP_TIMEOUT_INIT = 2
# SLEEP_TIMEOUT_MAX used by backoff and cater to mesos restriction
# https://github.com/apache/mesos/blob/0dc24dba9afdca1948eea5cad05f861162dc8dd6/src/slave/constants.hpp#L45
SLEEP_TIMEOUT_MAX = 15
SELECT_TIMEOUT = 2


class Connection(object):

    def __init__(self, addr, callback):
        host, port = addr.split(':', 2)
        port = int(port)
        self._addr = (host, port)
        self._sock = socket.socket()
        self._sock.setblocking(0)
        self.connected = False
        try:
            self._sock.connect(self._addr)
        except socket.error as e:
            if e.errno != errno.EAGAIN and e.errno != errno.EINPROGRESS:
                raise
        self._parser = HttpParser()
        self._callback = callback
        self._stream_id = None
        self._request = callback.gen_request()
        self._response = b''

    @property
    def addr(self):
        return self._addr

    @property
    def stream_id(self):
        return self._stream_id

    def handle_connect_event(self):
        err = self._sock.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
        if err == 0:
            self.connected = True
            return ""
        return _strerror(err)

    def write(self):
        try:
            sent = self._sock.send(self._request)
            self._request = self._request[sent:]
            return True
        except socket.error as e:
            if e.errno == errno.EAGAIN:
                return True

            logger.exception('Failed to send to %s', self._addr)
            return False

    def read(self):
        try:
            buf = self._sock.recv(PIPE_BUF)
            n_recv = len(buf)
            if n_recv == 0:
                logger.error('Remote %s closed', self.addr)
                return False

            n_parsed = self._parser.execute(buf, n_recv)
            if n_parsed != n_recv:
                if hasattr(self._parser, 'errno'):
                    # using http_parser.pyparser

                    from http_parser.pyparser import INVALID_CHUNK
                    if self._parser.errno == INVALID_CHUNK:
                        # need more chunk data
                        return True

                    if self._parser.errno:
                        raise RuntimeError(
                            'Failed to parse, code:%s %s' % (
                                self._parser.errno,
                                self._parser.errstr,
                            )
                        )

                else:
                    raise RuntimeError(
                        'Failed to parse, code:%s' % (
                            self._parser.get_errno(),
                        )
                    )

            if self._stream_id is None and self._parser.is_headers_complete():
                code = self._parser.get_status_code()
                if code == TEMPORARY_REDIRECT:
                    headers = {
                        k.upper(): v
                        for k, v in list(self._parser.get_headers().items())
                    }
                    new_master = headers['LOCATION']
                    new_master = urlparse(new_master).netloc or new_master
                    logger.warning(
                        'Try to redirect to new master: %s', new_master
                    )
                    self._callback.change_master(new_master)
                    return False

                elif code == SERVICE_UNAVAILABLE:
                    logger.warning('Master is not available, retry.')
                    return False

                elif code != OK:
                    msg = self._parser.recv_body()
                    if not self._parser.is_message_complete():
                        msg += ' ...'

                    raise RuntimeError('Failed with HTTP %s: %s' % (code, msg))
                if not self._parser.is_chunked():
                    raise RuntimeError('Response is not chunked')

                headers = {
                    k.upper(): v
                    for k, v in list(self._parser.get_headers().items())
                }
                self._stream_id = headers.get('MESOS-STREAM-ID', '')
                self._callback.stream_id = self._stream_id

            if self._parser.is_partial_body():
                self._response += self._parser.recv_body()
                while True:
                    m = LENGTH_PATTERN.match(self._response)
                    if not m:
                        break

                    captured = m.group(0)
                    length = int(captured.strip())
                    if len(self._response) < len(captured) + length:
                        break

                    data = self._response[
                        len(captured):len(captured) + length]
                    self._response = self._response[
                        len(captured) + length:]
                    try:
                        event = json.loads(data.decode('utf-8'))
                    except Exception:
                        logger.exception('Failed parse json %s', data)
                        raise

                    try:
                        self._callback.process_event(event)
                    except Exception:
                        logger.exception('Failed to process event')
                        raise

            if self._parser.is_message_complete():
                logger.debug('Event stream ended')
                return False

            return True
        except socket.error as e:
            if e.errno == errno.EAGAIN:
                return True

            logger.exception('Failed to recv from %s', self._addr)
            return False

    def want_write(self):
        return bool(self._request)

    def fileno(self):
        return self._sock.fileno()

    def close(self):
        self._sock.close()
        self._sock = None
        self.connected = False
        self._parser = None
        self._request = None
        self._response = None
        self._callback.on_close()


class Process(object):

    def __init__(self, master=None, timeout=DAY):
        self._master = None
        self._started = False
        self._lock = RLock()
        self._wakeup_fds = None
        self._io_thread = None
        self._new_master = master
        self._stream_id = None
        self._timeout = timeout

    @property
    def aborted(self):
        with self._lock:
            return not self._started

    @property
    def master(self):
        with self._lock:
            return self._master

    @property
    def stream_id(self):
        with self._lock:
            return self._stream_id

    @stream_id.setter
    def stream_id(self, _stream_id):
        with self._lock:
            self._stream_id = _stream_id

    @property
    def connected(self):
        return self.stream_id is not None

    def gen_request(self):
        raise NotImplementedError

    def on_event(self, event):
        raise NotImplementedError

    def on_close(self):
        raise NotImplementedError

    def process_event(self, event):
        if self._started:
            self.on_event(event)

    def change_master(self, new_master):
        with self._lock:
            self._new_master = new_master

        self._notify()

    def _notify(self):
        with self._lock:
            if self._wakeup_fds:
                os.write(self._wakeup_fds[1], b'\0')

    def _shutdown(self):
        pass

    def _run(self):
        try:
            with self._lock:
                _wakeup_fd = self._wakeup_fds[0]

            conn = None
            self.stream_id = None
            num_conn_retry = 0
            connect_deadline = 0

            while True:
                if not conn and self._master:
                    conn = Connection(self._master, self)

                    # reset deadline at first retry of every period
                    if num_conn_retry < 1:
                        connect_deadline = self._timeout + time.time()

                    if time.time() > connect_deadline:
                        raise Exception("connect reach deadline")

                to_write = set()
                to_read = set([_wakeup_fd])
                if conn:
                    to_read.add(conn.fileno())
                    if conn.want_write():
                        to_write.add(conn.fileno())

                readable, writeable, _ = select.select(to_read, to_write, [],
                                                       SELECT_TIMEOUT)

                if _wakeup_fd in readable:
                    with self._lock:
                        if not self._started:
                            break

                        if self._new_master != self._master:
                            if conn is not None:
                                to_read.discard(conn.fileno())
                                to_write.discard(conn.fileno())
                                conn.close()
                                conn = None
                                self.stream_id = None
                            self._master = self._new_master
                            num_conn_retry = 0

                    readable.remove(_wakeup_fd)
                    os.read(_wakeup_fd, PIPE_BUF)

                if not conn:
                    continue

                if not conn.connected:
                    err = conn.handle_connect_event()
                    if err:
                        sleep_timeout = self._backoff(num_conn_retry)
                        num_conn_retry += 1
                        conn.close()
                        conn = None

                        deadline = time.strftime('%Y-%m-%d %H:%M:%S',
                                                 time.localtime(
                                                     connect_deadline))
                        logger.warning(
                            'connect to %s error: %s,'
                            ' sleep %ds and try again,'
                            ' deadline: %s',
                            self._master, err, sleep_timeout, deadline)

                        time.sleep(sleep_timeout)
                        continue
                    else:
                        num_conn_retry = 0

                if writeable:
                    if not conn.write():
                        conn.close()
                        conn = None
                        self.stream_id = None
                        continue

                if readable:
                    if not conn.read():
                        conn.close()
                        conn = None
                        self.stream_id = None
        except Exception:
            logger.exception('Thread abort:')
            with self._lock:
                self._started = False

            global _exc_info
            _exc_info = sys.exc_info()
            thread.interrupt_main()

        finally:
            self._shutdown()

            if conn:
                conn.close()
                conn = None

            with self._lock:
                r, w = self._wakeup_fds
                os.close(r)
                os.close(w)
                self._wakeup_fds = None

    def _backoff(self, num_conn_retry):
        def _random_time(new, old):
            max_shift = (new - old) / SLEEP_TIMEOUT_SCALE
            return random.uniform(-max_shift, max_shift)

        new_timeout = SLEEP_TIMEOUT_INIT * math.pow(SLEEP_TIMEOUT_SCALE,
                                                    num_conn_retry)
        old_timeout = new_timeout / SLEEP_TIMEOUT_SCALE

        return min(SLEEP_TIMEOUT_MAX,
                   new_timeout + _random_time(new_timeout, old_timeout))

    def start(self):
        with self._lock:
            if self._started:
                logger.warning('Process already started!')
                return

        if self._io_thread:
            self.join()

        self._wakeup_fds = os.pipe()
        self._started = True
        self._io_thread = Thread(target=self._run, name='Process IO')
        self._io_thread.daemon = True
        self._io_thread.start()

    def abort(self):
        self.stop()

    def stop(self):
        with self._lock:
            self._started = False

        self._notify()

    def join(self):
        if self._io_thread:
            self._io_thread.join()
            self._io_thread = None

    def run(self):
        self.start()
        self.join()



import os
import re
import sys
from setuptools import setup, find_packages

needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)
pytest_runner = ['pytest-runner'] if needs_pytest else []


def find_version(*paths):
    fname = os.path.join(*paths)
    with open(fname) as fhandler:
        version_file = fhandler.read()
        version_match = re.search(r"^__VERSION__ = ['\"]([^'\"]*)['\"]",
                                  version_file, re.M)

    if not version_match:
        raise RuntimeError("Unable to find version string in %s" % (fname, ))

    version = version_match.group(1)
    return version


version = find_version('pymesos', '__init__.py')
install_requires=[
    'six',
    'http-parser',
    'addict',
]
PY3 = sys.version_info > (3, )
PYPY = getattr(sys, 'pypy_version_info', False) and True or False

if (PY3 or PYPY):
    install_requires += ['kazoo']
else:
    install_requires += ['zkpython']
print(install_requires)
setup(
    name='pymesos',
    version=version,
    description="A pure python implementation of Mesos scheduler and executor",
    packages=find_packages(),
    platforms=['POSIX'],
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX',
        'Programming Language :: Python',
    ],
    keywords='mesos',
    author="Zhongbo Tian",
    author_email="tianzhongbo@douban.com",
    url="https://github.com/douban/pymesos",
    download_url=('https://github.com/douban/pymesos/archive/%s.tar.gz' %
                  version),
    install_requires=install_requires,
    setup_requires=pytest_runner,
    tests_require=['pytest-cov', 'pytest-randomly', 'pytest-mock', 'pytest'],
)
